#pragma once

#include <lektor/common.h>
#include <sys/sysinfo.h>
#include <sched.h>
#include <pthread.h>
#include <stdlib.h>

typedef void *(*worker_pool_function)(void *, void *);

typedef enum {
    WORKER_STATUS_NONE      = 0,
    WORKER_STATUS_INTERRUPT = (1 << 1),
} WORKER_STATUS;

struct worker_pool {
    volatile size_t len;
    volatile size_t size;
    volatile size_t thread_size;
    volatile size_t thread_working;
    volatile worker_pool_function *functions;
    volatile void *volatile *volatile args;
    pthread_t *threads;
    pthread_mutex_t lock;
    volatile int exit;
};

/* After a free, the pool is no longer usable, you must re-init with a new */
PRIVATE_FUNCTION int worker_pool_new(struct worker_pool *, size_t init_len, size_t thread_count);
PRIVATE_FUNCTION void worker_pool_free(struct worker_pool *);
PRIVATE_FUNCTION void worker_pool_waitall(struct worker_pool *);

/* Request an interupt for all threads/tasks, you can wait after that. Note
 * that if a task is currently executed, it needs to supports a way of being
 * interupted (no automagic stuff here). */
PRIVATE_FUNCTION void worker_pool_interrupt(struct worker_pool *);

/* It is up to the user to free or not the `arg` depending on what he whats */
PRIVATE_FUNCTION int worker_pool_push(struct worker_pool *, worker_pool_function, void *arg);

/* Get the working thread count */
PRIVATE_FUNCTION size_t worker_pool_get_working_count(struct worker_pool *);

/* For clients, get the status of the worker */
PRIVATE_FUNCTION WORKER_STATUS worker_pool_get_status(void *);

/* Implementation */

PRIVATE_FUNCTION void *
worker_pool_thread(void *___pool)
{
    lkt_thread_set_name("lektord/worker");

    volatile struct worker_pool *pool = (volatile struct worker_pool *)___pool;
    volatile void *volatile arg;
    worker_pool_function func;
    FOR_EVER
    {
        lkt_assert(!pthread_mutex_lock((pthread_mutex_t *)&pool->lock));
        if (pool->len) {
            --(pool->len);
            ++(pool->thread_working);
            func = pool->functions[pool->len];
            arg  = pool->args[pool->len];
        } else if (pool->exit) {
            lkt_assert(!pthread_mutex_unlock((pthread_mutex_t *)&pool->lock));
            LOG_INFO("WORKER", "Exiting");
            break;
        } else {
            lkt_assert(!pthread_mutex_unlock((pthread_mutex_t *)&pool->lock));
            sleep(1);
            continue;
        }

        lkt_assert(!pthread_mutex_unlock((pthread_mutex_t *)&pool->lock));
        LOG_INFO("WORKER", "Picked up a function");
        func(___pool, (void *)arg);
        LOG_INFO("WORKER", "Finished work for a function");

        lkt_assert(!pthread_mutex_lock((pthread_mutex_t *)&pool->lock));
        --(pool->thread_working);
        lkt_assert(!pthread_mutex_unlock((pthread_mutex_t *)&pool->lock));
    }
    pthread_exit(NULL);
}

PRIVATE_FUNCTION WORKER_STATUS
worker_pool_get_status(void *__pool)
{
    WORKER_STATUS ret        = WORKER_STATUS_NONE;
    struct worker_pool *pool = (struct worker_pool *)__pool;

    if (pool == NULL) {
        LOG_WARN("WORKER", "Can't find the status of a NULL worker");
        return WORKER_STATUS_NONE;
    }

    /* Case handling to get the correct status, simple for the moment, but may
     * involve some logic in the future. */
    DO_WITH_LOCK(pool->lock, error, {
        if (pool->exit && pool->len == 0)
            ret = WORKER_STATUS_INTERRUPT;
    });

error:
    return ret;
}

PRIVATE_FUNCTION void
worker_pool_interrupt(struct worker_pool *pool)
{
    DO_WITH_LOCK(pool->lock, error, {
        pool->len  = 0;
        pool->exit = 1;
    });
error:
    return;
}

PRIVATE_FUNCTION int
worker_pool_new(struct worker_pool *ret, size_t init_size, size_t thread_count)
{
    if (!thread_count) {
        int nprocs = get_nprocs_conf();
        lkt_assert(nprocs > 0);
        thread_count = (size_t)nprocs;
    }
    struct worker_pool __ret = {
        .functions   = LKT_ALLOC_ARRAY(worker_pool_function, init_size),
        .args        = LKT_ALLOC_ARRAY(volatile void *, init_size),
        .threads     = LKT_ALLOC_ARRAY(pthread_t, thread_count),
        .size        = init_size,
        .thread_size = thread_count,
        .len         = 0u,
        .exit        = 0,
    };
    *ret = __ret;
    pthread_mutexattr_t attr;
    lkt_assert(!pthread_mutexattr_init(&attr));
    lkt_assert(!pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE));
    lkt_assert(!pthread_mutex_init(&ret->lock, &attr));
    lkt_assert(!pthread_mutexattr_destroy(&attr));
    size_t i;
    for (i = 0; i < ret->thread_size; ++i)
        lkt_assert(!pthread_create(&ret->threads[i], NULL, worker_pool_thread, ret));
    return 0;
}

PRIVATE_FUNCTION void
worker_pool_free(struct worker_pool *pool)
{
    LOG_DEBUG("WORKER", "Freeing worker pool %p", pool);
    size_t i;
    pool->exit = 1;
    for (i = 0; i < pool->thread_size; ++i)
        pthread_join(pool->threads[i], NULL);

    DO_WITH_LOCK(pool->lock, error, {
        safe_free((void **)&pool->threads);
        safe_free((void **)&pool->functions);
        safe_free((void **)&pool->args);
        pool->thread_size = 0u;
        pool->size        = 0u;
        pool->len         = 0u;
    });

error:
    return;
}

PRIVATE_FUNCTION int
worker_pool_push(struct worker_pool *pool, worker_pool_function func, void *arg)
{
    int ret = 1;

    DO_WITH_LOCK(pool->lock, error, {
        if (pool->exit) {
            LOG_ERROR("WORKER", "Can't push new jobs to worker if interupt was requested");
        }

        else {
            if (pool->len == pool->size) {
                if (pool->size < 1)
                    pool->size = 2;
                size_t nsize = 2 * pool->size;
                void *new_func =
                    safe_realloc((void *)pool->functions, nsize * sizeof(worker_pool_function));
                void *new_args = safe_realloc((void *)pool->args, nsize * sizeof(void *));

                pool->size *= 2;
                pool->functions = new_func;
                pool->args      = new_args;
            }

            pool->functions[pool->len] = func;
            pool->args[pool->len]      = arg;
            ++(pool->len);
            ret = 0;
        }
    });

error:
    return ret;
}

PRIVATE_FUNCTION void
worker_pool_waitall(struct worker_pool *pool)
{
    /* No lock, nothing, just test and yield */
    FOR_EVER_IF (pool->len)
        sched_yield();
}

PRIVATE_FUNCTION size_t
worker_pool_get_working_count(struct worker_pool *pool)
{
    return pool->thread_working;
}
