//! Create a connexion to a lektord instande then send commands and recieve data
//! from the server.

use crate::{
    constants, query::LektorQuery, query::LektorQueryLineType, response::LektorFormatedResponse,
    LektorResponse,
};
use log::*;
use std::{
    io::{self, BufRead, BufReader, Write},
    net::TcpStream,
    str::FromStr,
};

pub enum LektorQueryError {
    IO(io::Error),
    Query(String),
    Other(String),
    ToTyped(String),
}

impl std::fmt::Display for LektorQueryError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        use LektorQueryError::*;
        match self {
            IO(io) => write!(f, "lektor query error io: {io}"),
            Query(err) => write!(f, "lektor query logic error: {err}"),
            Other(other) => write!(f, "lektor query error: {other}"),
            ToTyped(msg) => write!(f, "lektor formated to typed error: {msg}"),
        }
    }
}

pub struct LektorConnexion {
    pub version: String,
    stream: TcpStream,
    reader: BufReader<TcpStream>,
}

impl std::fmt::Debug for LektorConnexion {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("LektorConnexion")
            .field("stream", &self.stream)
            .field("version", &self.version)
            .field("reader", &self.reader)
            .finish()
    }
}

impl LektorConnexion {
    pub fn new(hostname: impl AsRef<str>, port: i16) -> io::Result<Self> {
        match TcpStream::connect(format!("{}:{}", hostname.as_ref(), port)) {
            Ok(lektord) => {
                let mut ret: Self = Self {
                    stream: lektord.try_clone().unwrap(),
                    version: constants::MPD_VERSION.to_owned(),
                    reader: BufReader::new(lektord),
                };
                let mut daemon_version: String = String::new();
                ret.reader.read_line(&mut daemon_version).unwrap();
                assert_eq!(
                    daemon_version.trim(),
                    format!("OK MPD {}", ret.version),
                    "Expected MPD server version {}, but got version {}, abort!",
                    ret.version,
                    daemon_version
                );
                Ok(ret)
            }
            Err(e) => {
                error!("Failed to connect to lektor: {e}");
                Err(e)
            }
        }
    }

    pub fn send_query(&mut self, query: LektorQuery) -> Result<LektorResponse, LektorQueryError> {
        let mut res: Vec<String> = Vec::new();
        query.verify().map_err(LektorQueryError::Query)?;
        let builder = query.get_response_type();
        self.send_query_inner(query, &mut res)
            .map_err(LektorQueryError::IO)?;
        let formated = LektorFormatedResponse::try_from(res).map_err(LektorQueryError::Other)?;
        builder(formated).map_err(LektorQueryError::ToTyped)
    }

    fn send_query_inner(
        &mut self,
        query: LektorQuery,
        previous_ret: &mut Vec<String>,
    ) -> io::Result<()> {
        self.write_string(query.to_string())?;
        loop {
            match self.read_replies() {
                Err(e) => return Err(e),
                Ok((res, _)) if res.is_empty() => return Ok(()),
                Ok((res, None)) => return Ok(previous_ret.extend(res)),
                Ok((res, Some(cont))) => {
                    previous_ret.extend(res);
                    self.write_string(
                        LektorQuery::create_continuation(query.clone(), cont).to_string(),
                    )?;
                }
            }
        }
    }

    fn read_replies(&mut self) -> io::Result<(Vec<String>, Option<usize>)> {
        let error_return_value = io::Result::Err(io::Error::from(io::ErrorKind::Other));
        let mut ret: Vec<String> = Vec::new();
        let mut reply_line: String = String::new();
        let mut continuation = None;
        loop {
            reply_line.clear();
            match self.reader.read_line(&mut reply_line) {
                Err(e) => return io::Result::Err(e),
                Ok(size) => {
                    if size == 0 {
                        return io::Result::Err(io::Error::new(
                            io::ErrorKind::Other,
                            "recieved empty line",
                        ));
                    }
                    let msg = reply_line.trim();
                    match LektorQueryLineType::from_str(msg) {
                        Ok(LektorQueryLineType::Ok) => return Ok((ret, continuation)),
                        Ok(LektorQueryLineType::Ack) => return error_return_value,
                        Ok(LektorQueryLineType::Data) => ret.push(msg.to_string()),
                        Ok(LektorQueryLineType::ListOk) => continue,
                        Ok(LektorQueryLineType::Continuation(cont)) => continuation = Some(cont),
                        Err(_) => {
                            return Err(io::Error::new(
                                io::ErrorKind::Other,
                                "unknown query line type",
                            ))
                        }
                    }
                }
            }
        }
    }

    fn write_string(&mut self, buffer: impl AsRef<str>) -> io::Result<()> {
        let buffer = buffer.as_ref();
        if buffer.len() >= constants::LKT_MESSAGE_MAX {
            Err(io::Error::new(
                io::ErrorKind::InvalidData,
                "buffer to send is to big!",
            ))
        } else if !buffer.ends_with('\n') {
            Err(io::Error::new(
                io::ErrorKind::InvalidData,
                "buffer dosen't end with a new line",
            ))
        } else {
            self.stream.write_all(buffer.as_bytes())
        }
    }
}
